import logging
import socket
from types import MappingProxyType
from typing import Mapping  # NOQA

import yarl
from aiohttp import TCPConnector, ClientSession
from raven import Client
from raven.handlers.logging import SentryHandler
from raven.transport import Transport
from raven_aiohttp import QueuedAioHttpTransport

from aiomisc.service import Service

log = logging.getLogger(__name__)


class DummyTransport(Transport):
    def send(self, url, data, headers):
        pass


class QueuedKeepaliveAioHttpTransport(QueuedAioHttpTransport):
    DNS_CACHE_TTL = 600
    DNS_CACHE = True
    TCP_CONNECTION_LIMIT = 32
    TCP_CONNECTION_LIMIT_HOST = 8
    WORKERS = 1
    QUEUE_SIZE = 1000

    def __init__(self, *args, family=socket.AF_UNSPEC, loop=None,
                 dns_cache: bool = DNS_CACHE,
                 dns_cache_ttl: int = DNS_CACHE_TTL,
                 connection_limit: int = TCP_CONNECTION_LIMIT,
                 connection_limit_host: int = TCP_CONNECTION_LIMIT_HOST,
                 workers: int = WORKERS, qsize: int = QUEUE_SIZE, **kwargs):
        self.connection_limit = connection_limit
        self.connection_limit_host = connection_limit_host
        self.dns_cache = dns_cache
        self.dns_cache_ttl = dns_cache_ttl

        super().__init__(
            *args, family=family, loop=loop, keepalive=True,
            workers=workers, qsize=qsize, **kwargs
        )

    def _client_session_factory(self):
        self.connector = TCPConnector(
            family=self.family,
            limit=self.connection_limit,
            limit_per_host=self.connection_limit_host,
            loop=self._loop,
            ttl_dns_cache=self.dns_cache_ttl,
            use_dns_cache=self.dns_cache,
            verify_ssl=self.verify_ssl,
        )

        return ClientSession(
            connector=self.connector, loop=self._loop,
            connector_owner=False
        )

    async def _close(self):
        transport = await super()._close()
        self.connector.close()
        return transport


class RavenSender(Service):
    __required__ = 'sentry_dsn',

    sentry_dsn = None  # type: yarl.URL
    min_level = logging.WARNING  # type: int
    client_options = MappingProxyType({})  # type: Mapping

    client = None  # type: Client

    async def start(self):
        self.client = Client(
            str(self.sentry_dsn),
            transport=QueuedKeepaliveAioHttpTransport,
            **self.client_options
        )

        self.sentry_dsn = yarl.URL(
            self.sentry_dsn
        ).with_password('').with_user('')

        log.info('Starting Raven for %r', self.sentry_dsn)

        handler = SentryHandler(
            client=self.client,
            level=self.min_level
        )

        logging.getLogger().handlers.append(
            handler
        )

    async def stop(self, *_):
        transport = self.client.remote.get_transport()
        self.client.set_dsn('', transport=DummyTransport)

        await transport.close()

        log.info('Stopping Raven client for %r', self.sentry_dsn)


__all__ = ('RavenSender',)



import re
from types import MappingProxyType
from typing import List     # NOQA

from aiocarbon.protocol.pickle import PickleClient
from aiocarbon.protocol.tcp import TCPClient
from aiocarbon.protocol.udp import UDPClient
from aiocarbon.setup import set_client

from aiomisc.periodic import PeriodicCallback
from aiomisc.service import Service


def strip_carbon_ns(string):
    return re.sub(r'[^\w\d\-]+', '_', string).strip("_").lower()


PROTOCOLS = MappingProxyType({
    "udp": UDPClient,
    "tcp": TCPClient,
    "pickle": PickleClient,
})


class CarbonSender(Service):
    host = '127.0.0.1'          # type: str
    port = 2003                 # type: int
    send_interval = 5           # type: int
    protocol = 'udp'            # type: str
    namespace = ''              # type: List[str]
    _handle = None              # type: PeriodicCallback

    async def start(self):
        namespace = ".".join(
            strip_carbon_ns(item) for item in self.namespace
        )

        client = PROTOCOLS[self.protocol](
            self.host,
            self.port,
            namespace=namespace,
            loop=self.loop
        )

        set_client(client)

        self._handle = PeriodicCallback(client.send)
        self._handle.start(self.send_interval, loop=self.loop)

    async def stop(self, *_):
        self._handle.stop()



import asyncio
import socket
from functools import partial

from .base import SimpleServer
from ..utils import OptionsType, bind_socket


class UDPServer(SimpleServer):
    class UDPSimpleProtocol(asyncio.DatagramProtocol):

        def __init__(self, handle_datagram):
            super().__init__()
            self.handler = asyncio.coroutine(handle_datagram)
            self.transport = None  # type: asyncio.DatagramTransport
            self.loop = None  # type: asyncio.AbstractEventLoop

        def connection_made(self, transport: asyncio.DatagramTransport):
            self.transport = transport
            self.loop = asyncio.get_event_loop()

        def datagram_received(self, data: bytes, addr: tuple):
            self.loop.create_task(self.handler(data, addr))

    def __init__(self, address: str = None, port: int = None,
                 options: OptionsType = (), sock=None, **kwargs):
        if not sock:
            if not (address and port):
                raise RuntimeError(
                    'You should pass socket instance or '
                    '"address" and "port" couple'
                )

            self.make_socket = partial(
                bind_socket,
                socket.AF_INET6 if ':' in address else socket.AF_INET,
                socket.SOCK_DGRAM,
                address=address, port=port, options=options,
            )
        elif not isinstance(sock, socket.socket):
            raise ValueError('sock must be socket instance')
        else:
            self.make_socket = lambda: sock

        self.server = None
        self._protocol = None
        self.socket = None
        super().__init__(**kwargs)

    def handle_datagram(self, data: bytes, addr):
        raise NotImplementedError

    async def start(self):
        self.socket = self.make_socket()

        self.server, self._protocol = await self.loop.create_datagram_endpoint(
            lambda: UDPServer.UDPSimpleProtocol(self.handle_datagram),
            sock=self.socket,
        )



import tracemalloc
import logging
from ..periodic import PeriodicCallback
from ..service import Service
from ..thread_pool import threaded


log = logging.getLogger(__name__)


class MemoryTracer(Service):
    _tracer = None      # type: PeriodicCallback
    _log = None         # type: logging.Logger
    _snapshot_on_start = None

    logger = log.info

    interval = 5        # type: int
    top_results = 20    # type: int

    STAT_FORMAT = (
        "%(count)8s | "
        "%(count_diff)8s | "
        "%(size)8s | "
        "%(size_diff)8s | "
        "%(traceback)s\n"
    )

    async def start(self):
        tracemalloc.start()

        self._tracer = PeriodicCallback(self.show_stats)
        self._log = log.getChild(str(id(self)))

        self._snapshot_on_start = tracemalloc.take_snapshot()
        self._tracer.start(self.interval)

    @staticmethod
    def humanize(num, suffix='B'):
        for unit in ('', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi'):
            if abs(num) < 1024.0:
                return "%3.1f%s%s" % (num, unit, suffix)
            num /= 1024.0
        return "%.1f%s%s" % (num, 'Yi', suffix)

    @threaded
    def show_stats(self):
        snapshot = tracemalloc.take_snapshot()

        differences = snapshot.compare_to(self._snapshot_on_start, 'lineno')

        results = self.STAT_FORMAT % {
            "count": "Objects",
            "count_diff": "Obj.Diff",
            "size": "Memory",
            "size_diff": "Mem.Diff",
            "traceback": "Traceback"
        }
        for stat in differences[:self.top_results]:
            results += self.STAT_FORMAT % {
                "count": stat.count,
                "count_diff": stat.count_diff,
                "size": self.humanize(stat.size),
                "size_diff": self.humanize(stat.size_diff),
                "traceback": stat.traceback
            }

        self.logger("Top memory usage:\n%s", results)

    async def stop(self, exception: Exception = None):
        tracemalloc.stop()



import asyncio
from functools import partial

from .thread_pool import threaded


opener = threaded(open)


def async_method(name):
    async def method(self, *args, **kwargs):
        func = getattr(self.fp, name)

        return await self.loop.run_in_executor(
            None, partial(func, *args, **kwargs)
        )

    method.__name__ = name
    return method


# noinspection PyPep8Naming
class async_open:
    __slots__ = ('loop', 'name', 'mode', 'opener', 'fp', 'isatty', 'fileno')

    def __init__(self, fname, mode="r", *args, **kwargs):
        self.loop = kwargs.pop('loop', asyncio.get_event_loop())
        self.name = fname
        self.mode = mode
        self.opener = partial(opener, self.name, self.mode, *args, **kwargs)
        self.fp = None
        self.isatty = None
        self.fileno = None

    def closed(self):
        return self.fp.closed

    async def open(self):
        if self.fp is not None:
            return

        self.fp = await self.opener()
        self.fileno = self.fp.fileno()
        self.isatty = self.fp.isatty

    async def __aenter__(self):
        await self.open()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return await self.loop.run_in_executor(
            None, self.fp.__exit__, exc_type, exc_val, exc_tb
        )

    close = async_method('close')
    detach = async_method('detach')
    flush = async_method('flush')
    peek = async_method('peek')
    raw = async_method('raw')
    read = async_method('read')
    read1 = async_method('read1')
    readable = async_method('readable')
    readinto = async_method('readinto')
    readinto1 = async_method('readinto1')
    readline = async_method('readline')
    readlines = async_method('readlines')
    seek = async_method('seek')
    seekable = async_method('seekable')
    tell = async_method('tell')
    truncate = async_method('truncate')
    write = async_method('write')
    writable = async_method('writable')
    writelines = async_method('writelines')



import asyncio
import inspect
import time
from typing import AsyncIterator, Awaitable, Generator, TypeVar

import janus


T = TypeVar('T')
R = TypeVar('R')


class IteratorWrapper(AsyncIterator):
    def __init__(self, gen_func: Generator[T, None, R], loop=None, max_size=0):
        self.loop = loop or asyncio.get_event_loop()
        self.closed = False

        self.__close_event = asyncio.Event(loop=self.loop)
        self.__queue = janus.Queue(loop=self.loop, maxsize=max_size)
        self.__gen_task = self.loop.run_in_executor(
            None, self.__in_thread, gen_func
        )

    @staticmethod
    def __throw(_):
        pass

    def __in_thread(self, gen_func):
        queue = self.__queue.sync_q
        gen = iter(gen_func())

        throw = self.__throw
        if inspect.isgenerator(gen):
            throw = gen.throw

        try:
            while not self.closed:
                item = next(gen)

                while queue.full():
                    time.sleep(0)

                    if self.closed:
                        throw(asyncio.CancelledError())
                        return

                queue.put((item, False))
        except StopIteration as e:
            if self.closed:
                return
            queue.put((e, None))
        except Exception as e:
            if self.closed:
                return
            queue.put((e, True))
        finally:
            self.loop.call_soon_threadsafe(self.__close_event.set)

    async def close(self):
        self.closed = True
        self.__queue.close()
        await asyncio.gather(
            self.__queue.wait_closed(),
            return_exceptions=True
        )

        await self.__close_event.wait()

    async def __anext__(self) -> Awaitable[T]:
        item, is_exc = await self.__queue.async_q.get()
        self.__queue.async_q.task_done()

        if is_exc is None:
            await self.close()
            raise StopAsyncIteration(*item.args) from item
        elif is_exc:
            await self.close()
            raise item from item

        return item

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self.closed:
            return

        await self.close()



import asyncio
import logging
from contextlib import contextmanager
from typing import Tuple, Optional, Union

from .context import Context, get_context
from .log import basic_config, LogFormat
from .service import Service
from .utils import new_event_loop, select


def graceful_shutdown(services: Tuple[Service, ...],
                      loop: asyncio.AbstractEventLoop,
                      exception: Optional[Exception]):

    if hasattr(loop, '_default_executor'):
        try:
            loop._default_executor.shutdown()
        except Exception:
            logging.exception("Failed to stop default executor")

    tasks = [
        asyncio.shield(loop.create_task(svc.stop(exception)), loop=loop)
        for svc in services
    ]

    if not tasks:
        return

    loop.run_until_complete(
        asyncio.wait(
            tasks, loop=loop, return_when=asyncio.ALL_COMPLETED
        )
    )


@contextmanager
def entrypoint(*services: Service,
               loop: asyncio.AbstractEventLoop = None,
               pool_size: int = None,
               log_level: Union[int, str] = logging.INFO,
               log_format: Union[str, LogFormat] = 'color',
               log_buffer_size: int = 1024,
               log_flush_interval: float = 0.2,
               log_config: bool = True,
               debug: bool = False):
    """

    :param debug: set debug to event loop
    :param loop: loop
    :param services: Service instances which will be starting.
    :param pool_size: thread pool size
    :param log_level: Logging level which will be configured
    :param log_format: Logging format which will be configures
    :param log_buffer_size: Buffer size for logging
    :param log_flush_interval: interval in seconds for flushing logs
    :param log_config: if False do not configure logging
    """

    loop = loop or new_event_loop(pool_size)
    loop.set_debug(debug)

    ctx = Context(loop=loop)

    if log_config:
        basic_config(
            level=log_level,
            log_format=log_format,
            buffered=False,
        )

    async def start():
        nonlocal loop
        nonlocal services
        nonlocal log_format
        nonlocal log_level
        nonlocal log_buffer_size
        nonlocal log_flush_interval

        if log_config:
            basic_config(
                level=log_level,
                log_format=log_format,
                buffered=True,
                loop=loop,
                buffer_size=log_buffer_size,
                flush_interval=log_flush_interval,
            )

        starting = []

        async def start_service(svc: Service):
            await select(
                svc.start(), svc.start_event.wait(),
                cancel=False
            )

            if not svc.start_event.is_set():
                svc.start_event.set()

        for svc in services:
            svc.set_loop(loop)
            starting.append(loop.create_task(start_service(svc)))

        await asyncio.gather(*starting, loop=loop)

    loop.run_until_complete(start())

    shutting_down = False

    try:
        yield loop
    except Exception as e:
        graceful_shutdown(services, loop, e)
        shutting_down = True
        raise
    finally:
        if not shutting_down:
            graceful_shutdown(services, loop, None)

        loop.run_until_complete(loop.shutdown_asyncgens())
        ctx.close()


__all__ = ('entrypoint', 'get_context')



import asyncio
import logging
from functools import partial
from typing import Union
from . import utils


log = logging.getLogger(__name__)


class PeriodicCallback:
    """
    .. note::

        When the periodic function executes longer then execution interval a
        next call will be skipping and warning will be logged.

    """

    __slots__ = '_cb', '_closed', '_task', '_loop', '_handle', '__name'

    def __init__(self, coroutine_func, *args, **kwargs):
        self.__name = repr(coroutine_func)
        self._cb = partial(asyncio.coroutine(coroutine_func), *args, **kwargs)
        self._closed = False
        self._loop = None
        self._handle = None
        self._task = None

    async def _run(self):
        try:
            await self._cb()
        except Exception:
            log.exception("Periodic task error:")

    def start(self, interval: Union[int, float],
              loop=None, *, shield: bool = False):

        if self._closed:
            raise asyncio.InvalidStateError

        self._loop = loop or asyncio.get_event_loop()

        def periodic():
            if self._task and not self._task.done():
                log.warning('Task %r still running skipping', self)
                return

            self._task = self._loop.create_task(
                (utils.shield(self._run) if shield else self._run)()
            )

            if self._closed:
                return

            self._task.add_done_callback(call)

        def call(*_):
            self._handle = self._loop.call_later(interval, periodic)

        self._loop.call_soon_threadsafe(periodic)

    def stop(self):
        self._closed = True

        if self._task is None:
            self._task = asyncio.Future(loop=self._loop)
            self._task.set_exception(RuntimeError("Callback not started"))
        elif not self._task.done():
            self._task.cancel()

        if self._handle:
            self._handle.cancel()

        return self._task

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.__name)



import socket

from aiohttp.web import Application, AppRunner, SockSite
from aiomisc.service.tls import get_ssl_context, PathOrStr

try:
    from aiohttp.web_log import AccessLogger
except ImportError:         # pragma: nocover
    from aiohttp.helpers import AccessLogger

from .base import Service
from ..utils import bind_socket


class AIOHTTPService(Service):
    __async_required__ = 'start', 'create_application'

    def __init__(self, address: str = None, port: int = None,
                 sock: socket.socket = None, shutdown_timeout: int = 5, **kwds):

        if not sock:
            if not (address and port):
                raise RuntimeError(
                    'You should pass socket instance or '
                    '"address" and "port" couple'
                )

            self.socket = bind_socket(
                address=address,
                port=port,
                proto_name='http',
            )

        elif not isinstance(sock, socket.socket):
            raise ValueError('sock must be socket instance')
        else:
            self.socket = sock

        self.runner = None
        self.site = None
        self.shutdown_timeout = shutdown_timeout

        super().__init__(**kwds)

    async def create_application(self) -> Application:
        raise NotImplementedError('You should implement '
                                  '"create_application" method')

    async def create_site(self):
        return SockSite(
            self.runner, self.socket,
            shutdown_timeout=self.shutdown_timeout
        )

    async def start(self):
        self.runner = AppRunner(
            await self.create_application(),
            access_log_class=AccessLogger,
            access_log_format=AccessLogger.LOG_FORMAT,
        )

        await self.runner.setup()

        self.site = await self.create_site()

        await self.site.start()

    async def stop(self, exception: Exception = None):
        try:
            await self.site.stop()
        finally:
            await self.runner.cleanup()


class AIOHTTPSSLService(AIOHTTPService):
    def __init__(self, cert: PathOrStr, key: PathOrStr, ca: PathOrStr = None,
                 address: str = None, port: int = None, verify: bool = True,
                 sock: socket.socket = None, shutdown_timeout: int = 5,
                 require_client_cert: bool = False, **kwds):

        super().__init__(
            address=address, port=port, sock=sock,
            shutdown_timeout=shutdown_timeout, **kwds
        )

        self.__ssl_options = cert, key, ca, verify, require_client_cert

    async def create_site(self):
        return SockSite(
            self.runner, self.socket,
            shutdown_timeout=self.shutdown_timeout,
            ssl_context=await self.loop.run_in_executor(
                None, get_ssl_context, *self.__ssl_options
            )
        )



import asyncio
from functools import wraps
from typing import Optional, Type, TypeVar, Union

from .timeout import timeout


Number = Union[int, float]
T = TypeVar('T')


# noinspection SpellCheckingInspection
def asyncbackoff(attempt_timeout: Optional[Number],
                 deadline: Optional[Number],
                 pause: Number = 0, *exceptions: Type[Exception]):
    if not pause:
        pause = 0
    elif pause < 0:
        raise ValueError("'pause' must be positive")

    if attempt_timeout is not None and attempt_timeout < 0:
        raise ValueError("'attempt_timeout' must be positive or None")

    if deadline is not None and deadline < 0:
        raise ValueError("'deadline' must be positive or None")

    exceptions = tuple(exceptions) or ()
    exceptions += asyncio.TimeoutError,

    def decorator(func):
        if attempt_timeout is not None:
            func = timeout(attempt_timeout)(func)

        @wraps(func)
        async def wrap(*args, **kwargs):
            loop = asyncio.get_event_loop()
            last_exc = None

            async def run():
                nonlocal last_exc

                while True:
                    try:
                        return await asyncio.wait_for(
                            func(*args, **kwargs),
                            loop=loop,
                            timeout=attempt_timeout
                        )
                    except asyncio.CancelledError:
                        raise
                    except exceptions as e:
                        last_exc = e
                        await asyncio.sleep(pause, loop=loop)

            try:
                return await asyncio.wait_for(
                    run(), timeout=deadline, loop=loop
                )
            except Exception:
                if last_exc:
                    raise last_exc
                raise

        return wrap
    return decorator
